r"""Contain data structures to manage a collection of shards."""

from __future__ import annotations

__all__ = ["ShardDict", "ShardList"]

from iden.shard.collection.dict import ShardDict
from iden.shard.collection.list import ShardList
